@@ -36,6 +36,35 @@
 #include <gtk/gtk.h>
 #endif
 
+#if !GTK_CHECK_VERSION(2, 4, 0)
+enum {
+	SHOW_PREVIEW_LABEL = 1,
+	HIDE_PREVIEW_LABEL
+};
+#endif
+
+#if !GTK_CHECK_VERSION(2, 4, 0)
+static void
+_multiple_foreach(GtkTreeModel *model, GtkTreePath  *path, GtkTreeIter  *itr,
+		gpointer data)
+{
+	gchar **filename = data;
+	
+	if(!*filename)
+		gtk_tree_model_get(model, itr, 0, filename, -1);
+}
+
+static gchar *
+_get_label_from_treesel(GtkTreeSelection *selection)
+{
+	gchar *filename = NULL;
+	
+	gtk_tree_selection_selected_foreach(selection, _multiple_foreach, &filename);
+	
+	return filename;
+}
+#endif
+
 static void
 #if GTK_CHECK_VERSION(2, 4, 0)
 _preview_needs_update(GtkFileChooser *chooser, gpointer user_data)
@@ -47,6 +76,19 @@ _preview_needs_update(GtkTreeSelection *selection, gpointer user_data)
 	gpointer data = NULL;
 #if !GTK_CHECK_VERSION(2, 4, 0)
 	XfceFileChooser *chooser = XFCE_FILE_CHOOSER(user_data);
+	GtkWidget *frame = g_object_get_data(G_OBJECT(chooser), "xfce-fc-pframe");
+	
+	if(GPOINTER_TO_INT(g_object_get_data(G_OBJECT(chooser), "xfce-fc-plabel"))
+			== HIDE_PREVIEW_LABEL)
+	{
+		gtk_frame_set_label(GTK_FRAME(frame), NULL);
+	} else {
+		gchar *filename = _get_label_from_treesel(selection);
+		gtk_frame_set_label(GTK_FRAME(frame), filename);
+		if(filename)
+			g_free(filename);
+	}
+		
 #endif
 	
 	func = g_object_get_data(G_OBJECT(chooser), "xfce-fc-previewfunc");
@@ -56,6 +98,71 @@ _preview_needs_update(GtkTreeSelection *selection, gpointer user_data)
 		(*func)(XFCE_FILE_CHOOSER(chooser), data);
 }
 
+static XfceFileChooser *
+xfce_file_chooser_new_valist(const gchar *title, GtkWindow *parent,
+		XfceFileChooserAction action, const gchar *first_button_text,
+		va_list ap)
+{
+	XfceFileChooser *chooser;
+#if GTK_CHECK_VERSION(2, 4, 0)
+
+	chooser = XFCE_FILE_CHOOSER(gtk_file_chooser_dialog_new(title, parent,
+			action, NULL));
+	
+	if(first_button_text) {
+		const gchar *text = first_button_text;
+		GtkResponseType resp;
+		
+		while(text) {
+			resp = va_arg(ap, gint);
+			gtk_dialog_add_button(GTK_DIALOG(chooser), text, resp);
+			text = va_arg(ap, const gchar *);
+		}
+	}
+	
+	g_signal_connect(G_OBJECT(chooser), "update-preview",
+			G_CALLBACK(_preview_needs_update), NULL);
+#else
+	GtkTreeSelection *sel;
+	GtkWidget *action_area;
+	GList *buttons, *l;
+	
+	chooser = XFCE_FILE_CHOOSER(gtk_file_selection_new(title));
+
+	if(parent) {
+		gtk_window_set_transient_for(GTK_WINDOW(chooser), parent);
+		gtk_window_set_destroy_with_parent(GTK_WINDOW(chooser), TRUE);
+	}
+	gtk_dialog_set_has_separator(GTK_DIALOG(chooser), FALSE);
+	
+	/* i dislike just ditching the buttons gtk 2.2 gives us, but i want to make
+	 * sure the caller is getting the right responses that s/he expects */
+	action_area = GTK_DIALOG(chooser)->action_area;
+	buttons = gtk_container_get_children(GTK_CONTAINER(action_area));
+	for(l = buttons; l; l = l->next) {
+		if(l->data)
+			gtk_container_remove(GTK_CONTAINER(action_area), GTK_WIDGET(l->data));
+	}
+	
+	if(first_button_text) {
+		const gchar *text = first_button_text;
+		GtkResponseType resp;
+		
+		while(text) {
+			resp = va_arg(ap, gint);
+			gtk_dialog_add_button(GTK_DIALOG(chooser), text, resp);
+			text = va_arg(ap, const gchar *);
+		}
+	}
+	
+	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(GTK_FILE_SELECTION(chooser)->file_list));
+	g_signal_connect(G_OBJECT(sel), "changed",
+			G_CALLBACK(_preview_needs_update), chooser); 
+#endif
+	
+	return chooser;
+}
+
 /**
  * xfce_file_chooser_dialog_new:
  *
@@ -66,45 +173,12 @@ xfce_file_chooser_dialog_new(const gchar *title, GtkWindow *parent,
 		XfceFileChooserAction action, const gchar *first_button_text, ...)
 {
 	XfceFileChooser *chooser;
-#if GTK_CHECK_VERSION(2, 4, 0)
-
-	chooser = XFCE_FILE_CHOOSER(gtk_file_chooser_dialog_new(title, parent,
-			action, NULL));
+	va_list ap;
 	
-	if(first_button_text) {
-		va_list ap;
-		const gchar *text;
-		GtkResponseType resp;
-		
-		va_start(ap, first_button_text);
-		text = first_button_text;
-		while(text) {
-			resp = va_arg(ap, gint);
-			gtk_dialog_add_button(GTK_DIALOG(chooser), text, resp);
-			text = va_arg(ap, const gchar *);
-		}
-		va_end(ap);
-	}
-	
-	g_signal_connect(G_OBJECT(chooser), "update-preview",
-			G_CALLBACK(_preview_needs_update), NULL);
-#else
-	GtkTreeSelection *sel;
-	
-	chooser = XFCE_FILE_CHOOSER(gtk_file_selection_new(title));
-
-	if(parent) {
-		gtk_window_set_transient_for(GTK_WINDOW(chooser), parent);
-		gtk_window_set_destroy_with_parent(GTK_WINDOW(chooser), TRUE);
-	}
-	gtk_dialog_set_has_separator(GTK_DIALOG(chooser), FALSE);
-	
-	/* FIXME: do something with the buttons? */
-	
-	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(GTK_FILE_SELECTION(chooser)->file_list));
-	g_signal_connect(G_OBJECT(sel), "changed",
-			G_CALLBACK(_preview_needs_update), chooser); 
-#endif
+	va_start(ap, first_button_text);
+	chooser = xfce_file_chooser_new_valist(title, parent, action,
+			first_button_text, ap);
+	va_end(ap);
 	
 	return GTK_WIDGET(chooser);
 }
@@ -134,45 +208,12 @@ xfce_file_chooser_new(const gchar *title, GtkWindow *parent,
 		XfceFileChooserAction action, const gchar *first_button_text, ...)
 {
 	XfceFileChooser *chooser;
-#if GTK_CHECK_VERSION(2, 4, 0)
-
-	chooser = XFCE_FILE_CHOOSER(gtk_file_chooser_dialog_new(title, parent,
-			action, NULL));
+	va_list ap;
 	
-	if(first_button_text) {
-		va_list ap;
-		const gchar *text;
-		GtkResponseType resp;
-		
-		va_start(ap, first_button_text);
-		text = first_button_text;
-		while(text) {
-			resp = va_arg(ap, gint);
-			gtk_dialog_add_button(GTK_DIALOG(chooser), text, resp);
-			text = va_arg(ap, const gchar *);
-		}
-		va_end(ap);
-	}
-	
-	g_signal_connect(G_OBJECT(chooser), "update-preview",
-			G_CALLBACK(_preview_needs_update), NULL);
-#else
-	GtkTreeSelection *sel;
-	
-	chooser = XFCE_FILE_CHOOSER(gtk_file_selection_new(title));
-
-	if(parent) {
-		gtk_window_set_transient_for(GTK_WINDOW(chooser), parent);
-		gtk_window_set_destroy_with_parent(GTK_WINDOW(chooser), TRUE);
-	}
-	gtk_dialog_set_has_separator(GTK_DIALOG(chooser), FALSE);
-	
-	/* FIXME: do something with the buttons? */
-	
-	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(GTK_FILE_SELECTION(chooser)->file_list));
-	g_signal_connect(G_OBJECT(sel), "changed",
-			G_CALLBACK(_preview_needs_update), chooser); 
-#endif
+	va_start(ap, first_button_text);
+	chooser = xfce_file_chooser_new_valist(title, parent, action,
+			first_button_text, ap);
+	va_end(ap);
 	
 	return GTK_WIDGET(chooser);
 }
@@ -382,7 +423,7 @@ xfce_file_chooser_get_current_folder(XfceFileChooser *chooser)
 	
 	dir = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(chooser)));
 	if(dir) {
-		if(p=g_strrstr(dir, G_DIR_SEPARATOR_S))
+		if((p=g_strrstr(dir, G_DIR_SEPARATOR_S)))
 			*p = 0;
 	}
 	
@@ -412,14 +453,19 @@ xfce_file_chooser_set_preview_widget(XfceFileChooser *chooser,
 	gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(chooser), preview_widget);
 #else
 	/* following code based loosely on Xfce's PreviewFileSelection widget */
-	GtkWidget *hbox, *frame = NULL, *old_widget;
+	GtkWidget *hbox, *frame = NULL, *old_widget, *label;
 
 	old_widget = g_object_get_data(G_OBJECT(chooser), "xfce-fc-pwidget");
+	label = g_object_get_data(G_OBJECT(chooser), "xfce-fc-plabel");
 	if(old_widget) {
 		frame = g_object_get_data(G_OBJECT(chooser), "xfce-fc-pframe");
 		gtk_widget_hide_all(old_widget);
 		gtk_container_remove(GTK_CONTAINER(frame), old_widget);
 		gtk_widget_destroy(old_widget);
+		if(label) {
+			gtk_container_remove(GTK_CONTAINER(frame), label);
+			gtk_widget_destroy(label);
+		}
 	}
 	
 	if(!frame) {
@@ -429,11 +475,12 @@ xfce_file_chooser_set_preview_widget(XfceFileChooser *chooser,
 		}
 		
 		if(!hbox) {
-			g_warning("%s: Can't find GtkFileSelection's HBox, unable to set preview widget");
+			g_warning("Can't find GtkFileSelection's HBox, unable to set preview widget");
 			return;
 		}
 		
 		frame = gtk_frame_new(NULL);
+		gtk_frame_set_label_align(GTK_FRAME(frame), 0.5, 0.5);
 		gtk_box_pack_end(GTK_BOX(hbox), frame, FALSE, FALSE, 0);
 		
 		g_object_set_data(G_OBJECT(chooser), "xfce-fc-pframe", frame);
@@ -560,7 +607,8 @@ xfce_file_chooser_set_use_preview_label(XfceFileChooser *chooser,
 #if GTK_CHECK_VERSION(2, 4, 0)
 	gtk_file_chooser_set_use_preview_label(GTK_FILE_CHOOSER(chooser), use_label);
 #else
-	/* FIXME: implement me! */
+	g_object_set_data(G_OBJECT(chooser), "xfce-fc-plabel",
+			GINT_TO_POINTER(use_label ? SHOW_PREVIEW_LABEL : HIDE_PREVIEW_LABEL));
 #endif
 }
 
@@ -572,10 +620,6 @@ xfce_file_chooser_set_use_preview_label(XfceFileChooser *chooser,
  * Sets a widget to display in the file chooser to provide extra options to the
  * user.
  *
- * Note: This currently does nothing when libxfcegui4 is compiled against GTK+
- * 2.2 or lower.  As such, it should not be relied upon for essential
- * functionality.
- *
  * Since 4.1
  **/
 void
@@ -585,8 +629,22 @@ xfce_file_chooser_set_extra_widget(XfceFileChooser *chooser,
 #if GTK_CHECK_VERSION(2, 4, 0)
 	gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(chooser), extra_widget);
 #else
-	/* FIXME: implement me! */
-	/* stub just so we don't crash on a get */
+	GtkWidget *old, *box;
+	
+	box = gtk_widget_get_parent(GTK_FILE_SELECTION(chooser)->selection_entry);
+	old = g_object_get_data(G_OBJECT(chooser), "xfce-fc-ewidget");
+	
+	if(old == extra_widget)
+		return;
+	if(old) {
+		if(GTK_IS_BOX(box))
+			gtk_container_remove(GTK_CONTAINER(box), old);
+		gtk_widget_destroy(old);
+	}
+	
+	if(extra_widget && GTK_IS_BOX(box))
+		gtk_box_pack_start(GTK_BOX(box), extra_widget, FALSE, FALSE, 0);
+	
 	g_object_set_data(G_OBJECT(chooser), "xfce-fc-ewidget", extra_widget);
 #endif
 }
@@ -607,7 +665,6 @@ xfce_file_chooser_get_extra_widget(XfceFileChooser *chooser)
 #if GTK_CHECK_VERSION(2, 4, 0)
 	return gtk_file_chooser_get_extra_widget(GTK_FILE_CHOOSER(chooser));
 #else
-	/* FIXME: implement me! */
 	return g_object_get_data(G_OBJECT(chooser), "xfce-fc-ewidget");
 #endif
 }
@@ -690,6 +747,8 @@ xfce_file_chooser_set_filter(XfceFileChooser *chooser, XfceFileFilter *filter)
  * Sets whether or not the file chooser can display and/or allow the selection
  * of non-local files.
  *
+ * Deprecated: This function isn't really useful in the context of Xfce.
+ *
  * Since 4.1
  **/
 void
@@ -710,6 +769,8 @@ xfce_file_chooser_set_local_only(XfceFileChooser *chooser, gboolean local_only)
  * Gets whether or not the file chooser can display and/or allow the selection
  * of non-local files.
  *
+ * Deprecated: This function isn't really useful in the context of Xfce.
+ *
  * Since 4.1
  *
  * Return value: %TRUE if non-local files are allowed, %FALSE otherwise.
@@ -734,6 +795,9 @@ xfce_file_chooser_get_local_only(XfceFileChooser *chooser)
  * Adds a folder to be displayed in the shortcut folders list in the file
  * chooser.  The folder name is validated, so only valid paths may be added.
  *
+ * Note: This currently does nothing when libxfcegui4 is compiled against GTK+
+ * 2.2 or lower.
+ *
  * Since 4.1
  *
  * Return value: %TRUE if the folder given is valid, %FALSE otherwise.  If
@@ -760,6 +824,9 @@ xfce_file_chooser_add_shortcut_folder(XfceFileChooser *chooser,
  *
  * Removes the specified folder from the file chooser's shortcut folder list.
  *
+ * Note: This currently does nothing when libxfcegui4 is compiled against GTK+
+ * 2.2 or lower.
+ *
  * Since 4.1
  *
  * Return value: %TRUE if the removal succeeded, %FALSE otherwise.  If @error is
@@ -831,7 +898,7 @@ xfce_file_filter_set_name(XfceFileFilter *filter, const gchar *name)
  * Adds a MIME type to the file filter.  All files matching this MIME type will
  * be displayed in the file chooser when the filter is selected.
  *
- * Note: this will always return %NULL when libxfcegui4 is compiled against GTK+
+ * Note: this function has no effect when libxfcegui4 is compiled against GTK+
  * 2.2 or below, as file filters are not supported.
  *
  * Since 4.1
@@ -853,7 +920,7 @@ xfce_file_filter_add_mime_type(XfceFileFilter *filter, const gchar *mime_type)
  * the pattern will be displayed in the file chooser when the filter is
  * selected.
  *
- * Note: this will always return %NULL when libxfcegui4 is compiled against GTK+
+ * Note: this function has no effect when libxfcegui4 is compiled against GTK+
  * 2.2 or below, as file filters are not supported.
  *
  * Since 4.1
